import { Observable, Subject } from 'rxjs';

import { RPCClientError } from '../protocol/RPCError';
import { RPCClientRWC } from './RPCClientRWC';

import {
  RPCClientCodec,
  RPCClientResponseCodec,
  RPCClientNotificationCodec,
} from '../protocol/RPCClientCodec';
import { RPCMessage } from '../core/type';

export interface RPCRequestState {
  subject: Subject<any>;
  request: {
    method: string;
    params: any[];
  };
}

export abstract class RPCClient {
  private requestID: number;

  private pendingRequestsCount: number;
  private pendingRequests: Map<number, RPCRequestState>;

  protected rpcClientCodec: RPCClientCodec;
  protected rpcClientRWC: RPCClientRWC;

  public constructor(
    rpcClientCodec: RPCClientCodec,
    rpcClientRWC: RPCClientRWC,
  ) {
    this.rpcClientCodec = rpcClientCodec;
    this.rpcClientRWC = rpcClientRWC;
    this.requestID = 0;
    this.pendingRequestsCount = 0;
    this.pendingRequests = new Map();
  }

  private getRequestID(): number {
    return ++this.requestID;
  }

  /**
   * connect
   */
  public connect(queryString?: string): void {
    this.rpcClientRWC.connect(queryString);
    this.rpcClientRWC.readResponse().subscribe(
      (value: RPCMessage) => {
        this.onMessage(value);
      },
      (error: any) => {
        console.error(error);
      },
      () => {

      }
    );
  }

  /**
   * close
   */
  public disconnect() {
    this.rpcClientRWC.disconnect();
  }

  /**
   * notify
   */
  public send(method: string, ...args: any[]): void {
    this.sendInternal(false, method, args);
  }

  /**
   * call
   */
  public call<RT>(method: string, ...args: any[]): Observable<RT> {
    return this.sendInternal<RT>(true, method, args);
  }

  /**
   * callTimeout
   */
  public callTimeout<RT>(ms: number, method: string, ...args: any[]): Observable<RT> {

    return undefined;
  }

  private sendInternal<RT>(hasResponse: boolean, method: string, args?: any[]): Observable<RT> | undefined {
    let id: number;
    let resSubject: Subject<RT>;
    if (hasResponse) {
      id = this.getRequestID();
      resSubject = new Subject<RT>();
      const reqState: RPCRequestState = {
        subject: resSubject,
        request: {
          method: method,
          params: args,
        }
      };
      this.pendingRequests.set(id, reqState);
      this.pendingRequestsCount++;
    }

    this.rpcClientRWC.writeRequest(this.rpcClientCodec.request(method, args, id));

    if (undefined !== resSubject) {
      return resSubject.asObservable();
    }
    return undefined;
  }

  private onMessage(message: RPCMessage): void {
    const resCodec = this.rpcClientCodec.response(message);

    if (resCodec.isNotification()) {
      this.onNotification(resCodec.notification());
    } else {
      this.onResponse(resCodec);
    }
  }

  protected onResponse(resCodec: RPCClientResponseCodec): void {
    const id = resCodec.id();
    const result = resCodec.result();
    const error = resCodec.error();

    const reqState: RPCRequestState = this.pendingRequests.get(id);

    this.pendingRequests.delete(id);
    this.pendingRequestsCount--;

    if (undefined !== error) {
      const rpcClientError: RPCClientError = {
        request: reqState.request,
        response: error,
      };
      console.error(rpcClientError);
      reqState.subject.error(rpcClientError);
    } else {
      reqState.subject.next(result);
    }
  }

  protected abstract onNotification(notiCodec: RPCClientNotificationCodec): void;
}
